#include "custom.h"
#include <string.h>
#include <stdio.h>

namespace CustomMatrices {

static void ReadMatrice(uint8_t *matrices, const char *coeffs, int nCoefficients)
{
   for ( int i = 0; i < nCoefficients - 1; i++ )
   {
      char number[256];
      
      strncpy(number, coeffs, strchr(coeffs, ' ') - coeffs);
      number[strchr(coeffs, ' ') - coeffs] = 0;
      matrices[i] = atoi(number);
      if ( matrices[i] == 0 ) matrices[i] = 1;
      coeffs = strchr(coeffs, ' ') + 1;
   }
   matrices[nCoefficients - 1] = atoi(coeffs);
   if ( matrices[nCoefficients - 1] == 0 ) matrices[nCoefficients - 1] = 1;
}

Filter::Filter(PClip child, const char *pszFileName, const char *luma4x4, const char *chroma4x4, const char *luma8x8,
                       const char *pluma4x4, const char *pchroma4x4, const char *pluma8x8, 
                       bool isInter, bool is8x8, int nQuantizer, bool isDeblocking, int nDeblockingStrength,
                       bool showSize, bool showPsnr, IScriptEnvironment *env) :
GenericVideoFilter(child), showSize(showSize), showPsnr(showPsnr)
{
   nWidth = vi.width;
   nHeight = vi.height;

   /* initialize x264 librairy */
   x264_param_t params;
   memset( &params, 0, sizeof(params) );

   /* fills in the custom matrices */
   ReadMatrice(params.cqm_4ic, chroma4x4, 16);
   ReadMatrice(params.cqm_4iy, luma4x4, 16);
   ReadMatrice(params.cqm_8iy, luma8x8, 64);
   ReadMatrice(params.cqm_4pc, pchroma4x4, 16);
   ReadMatrice(params.cqm_4py, pluma4x4, 16);
   ReadMatrice(params.cqm_8py, pluma8x8, 64);

   params.psz_cqm_file = strdup(pszFileName);

   /* fills the rest of the params */
   params.i_width = nWidth;
   params.i_height = nHeight;
   params.i_csp = X264_CSP_I420;
   params.i_frame_total = 0;
   params.i_fps_num = 1;
   params.i_fps_den = 1;

   params.analyse.b_psnr = showPsnr;

   params.i_keyint_min = 1;
   if ( !isInter )
   {
      params.i_keyint_max = 1; /* only iframes */
      params.i_scenecut_threshold = -1;
   }
   else
   {
      params.i_keyint_max = 250;
      params.i_keyint_min = 25;
      params.i_scenecut_threshold = 40;
   }

   params.b_deblocking_filter = isDeblocking;
   params.i_deblocking_filter_alphac0 = nDeblockingStrength;
   params.i_deblocking_filter_beta = nDeblockingStrength;

   params.i_cqm_preset = X264_CQM_CUSTOM;

   params.rc.i_qp_constant = nQuantizer;
   if ( !isInter )
      params.rc.f_ip_factor = 1.0f;
   else
      params.rc.f_ip_factor = 1.4f;

   params.rc.f_rate_tolerance = 0.1f;

   params.analyse.intra = X264_ANALYSE_I4x4;
   if ( is8x8 )
      params.analyse.intra |= X264_ANALYSE_I8x8;
   params.analyse.b_transform_8x8 = is8x8;

   if ( isInter )
   {
      params.analyse.inter = X264_ANALYSE_I4x4 | X264_ANALYSE_I8x8 | X264_ANALYSE_PSUB16x16 | X264_ANALYSE_PSUB8x8;
      params.analyse.i_me_method = X264_ME_DIA;
      params.analyse.i_me_range = 16;
      params.analyse.i_subpel_refine = 3;
      params.analyse.i_mv_range = 512;
   }

   x264 = x264_encoder_open( &params );

   if ( !x264 )
      env->ThrowError("AVCMatrices : unvalid settings");

   x264_picture_alloc( &output, X264_CSP_I420, nWidth, nHeight );
   x264_picture_alloc( &input, X264_CSP_I420, nWidth, nHeight );

   nFrames = 0;
}

Filter::~Filter()
{
   x264_picture_clean( &output );
   x264_picture_clean( &input );
   x264_encoder_close( x264 );
}

static void ConvertToX264Picture(const PVideoFrame &frame, int nWidth, int nHeight, x264_picture_t &picture, IScriptEnvironment* env)
{
   env->BitBlt( picture.img.plane[0], picture.img.i_stride[0],
      frame->GetReadPtr(PLANAR_Y), frame->GetPitch(PLANAR_Y), nWidth, nHeight);
   env->BitBlt( picture.img.plane[1], picture.img.i_stride[1],
      frame->GetReadPtr(PLANAR_U), frame->GetPitch(PLANAR_U), nWidth >> 1, nHeight >> 1);
   env->BitBlt( picture.img.plane[2], picture.img.i_stride[2],
      frame->GetReadPtr(PLANAR_V), frame->GetPitch(PLANAR_V), nWidth >> 1, nHeight >> 1);
}

static void ConvertToVideoFrame(PVideoFrame &frame, int nWidth, int nHeight, x264_picture_t &picture, IScriptEnvironment* env)
{
   env->BitBlt( frame->GetWritePtr(PLANAR_Y), frame->GetPitch(PLANAR_Y),
      picture.img.plane[0], picture.img.i_stride[0], nWidth, nHeight);
   env->BitBlt( frame->GetWritePtr(PLANAR_U), frame->GetPitch(PLANAR_U),
      picture.img.plane[1], picture.img.i_stride[1], nWidth >> 1, nHeight >> 1);
   env->BitBlt( frame->GetWritePtr(PLANAR_V), frame->GetPitch(PLANAR_V),
      picture.img.plane[2], picture.img.i_stride[2], nWidth >> 1, nHeight >> 1);
}

#include "info.h"

PVideoFrame __stdcall Filter::GetFrame(int n, IScriptEnvironment* env)
{
   PVideoFrame dst = env->NewVideoFrame(vi);
   PVideoFrame src = child->GetFrame(n, env);

   x264_nal_t *pNals;
   int nNals;

   ConvertToX264Picture( src, nWidth, nHeight, input, env );
   input.i_pts = nFrames++;

   x264_encoder_encode( x264, &pNals, &nNals, &input, &output );

   ConvertToVideoFrame( dst, nWidth, nHeight, output, env );

   int nSize = 0;
   for ( int i = 0; i < nNals; i++ )
      nSize += pNals[i].i_payload;

   free( pNals );

   char typesize[100];
   char typepsnr[100];
   if ( showPsnr )
      sprintf(typepsnr, "type : %s, psnr : luma : %7.4f, chroma : %7.4f, frame : %7.4f", 
         output.i_type == X264_TYPE_P ? "P" : "I",
         output.f_psnr[0], output.f_psnr[1], output.f_psnr[2]);
   if ( showSize )
      sprintf(typesize, "type : %s, size : %5i bytes",
         output.i_type == X264_TYPE_P ? "P" : "I", nSize );

   int y = 0;
   if ( showSize )
      DrawString(dst, 0, 0, typesize), y += 1;
   if ( showPsnr )
      DrawString(dst, 0, y, typepsnr);

   return dst;
}

}

AVSValue __cdecl CreateCustomMatrices(AVSValue args, void *user_data, IScriptEnvironment* env)
{
   return new CustomMatrices::Filter(
      args[0].AsClip(),
      args[1].Defined() ? args[1].AsString() : NULL,
      args[2].AsString("16 16 16 16 "
                       "16 16 16 16 "
                       "16 16 16 16 "
                       "16 16 16 16"),
      args[3].AsString("16 16 16 16 "
                       "16 16 16 16 "
                       "16 16 16 16 "
                       "16 16 16 16"),
      args[4].AsString("16 16 16 16 16 16 16 16 "
                       "16 16 16 16 16 16 16 16 "
                       "16 16 16 16 16 16 16 16 "
                       "16 16 16 16 16 16 16 16 "
                       "16 16 16 16 16 16 16 16 "
                       "16 16 16 16 16 16 16 16 "
                       "16 16 16 16 16 16 16 16 "
                       "16 16 16 16 16 16 16 16"),
      args[5].AsString("16 16 16 16 "
                       "16 16 16 16 "
                       "16 16 16 16 "
                       "16 16 16 16"),
      args[6].AsString("16 16 16 16 "
                       "16 16 16 16 "
                       "16 16 16 16 "
                       "16 16 16 16"),
      args[7].AsString("16 16 16 16 16 16 16 16 "
                       "16 16 16 16 16 16 16 16 "
                       "16 16 16 16 16 16 16 16 "
                       "16 16 16 16 16 16 16 16 "
                       "16 16 16 16 16 16 16 16 "
                       "16 16 16 16 16 16 16 16 "
                       "16 16 16 16 16 16 16 16 "
                       "16 16 16 16 16 16 16 16"),
      args[5].Defined() || args[6].Defined() || args[7].Defined() || args[1].Defined(),
      args[4].Defined() || args[7].Defined(),
      args[8].AsInt(26),
      args[9].Defined(),
      args[9].AsInt(0),
      args[10].AsBool(true),
      args[11].AsBool(true),
      env);
}

extern "C" __declspec(dllexport) const char* __stdcall AvisynthPluginInit2(IScriptEnvironment* env)
{
   env->AddFunction("AVCMatrices", "c[file]s[luma4x4]s[chroma4x4]s[luma8x8]s[pluma4x4]s[pchroma4x4]s[pluma8x8]s[quant]i[deblocking]i[size]b[psnr]b", CreateCustomMatrices, NULL);

   return("AVCMatrices");
}


